@@ -4,6 +4,7 @@
 """Defines the RAITextInsights class."""
 
 import json
+import logging
 import pickle
 import warnings
 from enum import Enum
@@ -12,36 +13,94 @@ from typing import Any, List, Optional, Union
 
 import numpy as np
 import pandas as pd
-from erroranalysis._internal.cohort_filter import FilterDataWithCohortFilters
 from ml_wrappers import wrap_model
+
+from erroranalysis._internal.cohort_filter import FilterDataWithCohortFilters
 from raiutils.data_processing import convert_to_list, serialize_json_safe
-from raiutils.models import ModelTask as RAIModelTask
 from raiutils.models import SKLearn, is_classifier
 from responsibleai._interfaces import Dataset, RAIInsightsData
 from responsibleai._internal.constants import (ManagerNames, Metadata,
                                                SerializationAttributes)
 from responsibleai.exceptions import UserConfigValidationException
+from responsibleai.feature_metadata import FeatureMetadata
 from responsibleai.rai_insights.rai_base_insights import RAIBaseInsights
-
 from responsibleai_text.common.constants import ModelTask
 from responsibleai_text.managers.error_analysis_manager import \
     ErrorAnalysisManager
 from responsibleai_text.managers.explainer_manager import ExplainerManager
-from responsibleai_text.utils.feature_extractors import extract_features
+from responsibleai_text.utils.feature_extractors import (extract_features,
+                                                         get_text_columns)
+
+module_logger = logging.getLogger(__name__)
+module_logger.setLevel(logging.INFO)
+
+try:
+    import evaluate
+except ImportError:
+    module_logger.debug(
+        'Could not import evaluate, required if using a QA model')
 
 _PREDICTIONS = 'predictions'
 _PREDICT_OUTPUT = 'predict_output'
-_TRAIN = 'train'
+_TEST = 'test'
 _TARGET_COLUMN = 'target_column'
 _TASK_TYPE = 'task_type'
 _CLASSES = 'classes'
 _META_JSON = Metadata.META_JSON
-_TRAIN_LABELS = 'train_labels'
 _JSON_EXTENSION = '.json'
 _PREDICT = 'predict'
 _PREDICT_PROBA = 'predict_proba'
 _EXT_TEST = '_ext_test'
 _EXT_FEATURES = '_ext_features'
+_FEATURE_METADATA = Metadata.FEATURE_METADATA
+_IDENTITY_FEATURE_NAME = 'identity_feature_name'
+_DATETIME_FEATURES = 'datetime_features'
+_TIME_SERIES_ID_FEATURES = 'time_series_id_features'
+_CATEGORICAL_FEATURES = 'categorical_features'
+_DROPPED_FEATURES = 'dropped_features'
+_QUESTION_TYPE = 'question_type'
+_TEXT_COLUMN = '_text_column'
+
+
+def _feature_metadata_from_dict(feature_meta_dict):
+    """Create a FeatureMetadata from a dictionary.
+
+    :param feature_meta_dict: The dictionary to create the FeatureMetadata
+        from.
+    :type feature_meta_dict: dict
+    :return: The FeatureMetadata created from the dictionary.
+    :rtype: FeatureMetadata
+    """
+    return FeatureMetadata(
+        identity_feature_name=feature_meta_dict[_IDENTITY_FEATURE_NAME],
+        datetime_features=feature_meta_dict[_DATETIME_FEATURES],
+        time_series_id_features=feature_meta_dict[_TIME_SERIES_ID_FEATURES],
+        categorical_features=feature_meta_dict[_CATEGORICAL_FEATURES],
+        dropped_features=feature_meta_dict[_DROPPED_FEATURES])
+
+
+def _add_extra_metadata_features(task_type, feature_metadata):
+    """Add extra metadata features for the given task type.
+
+    For question answering task, adds the question type feature.
+
+    :param task_type: The task type.
+    :type task_type: str
+    :param feature_metadata: The feature metadata.
+    :type feature_metadata: FeatureMetadata
+    :return: The feature metadata with extra metadata features added.
+    :rtype: FeatureMetadata
+    """
+    is_qa = task_type == ModelTask.QUESTION_ANSWERING
+    categorical_features = feature_metadata.categorical_features
+    is_cat_empty = categorical_features is None
+    if is_qa and (is_cat_empty or _QUESTION_TYPE not in categorical_features):
+        feature_metadata = _feature_metadata_from_dict(
+            feature_metadata.to_dict().copy())
+        if is_cat_empty:
+            feature_metadata.categorical_features = []
+        feature_metadata.categorical_features.append(_QUESTION_TYPE)
+    return feature_metadata
 
 
 class RAITextInsights(RAIBaseInsights):
@@ -51,19 +110,19 @@ class RAITextInsights(RAIBaseInsights):
     single API.
     """
 
-    def __init__(self, model: Any, train: pd.DataFrame, test: pd.DataFrame,
+    def __init__(self, model: Any, test: pd.DataFrame,
                  target_column: str, task_type: str,
                  classes: Optional[np.ndarray] = None,
                  serializer: Optional[Any] = None,
-                 maximum_rows_for_test: int = 5000):
+                 maximum_rows_for_test: int = 5000,
+                 feature_metadata: Optional[FeatureMetadata] = None,
+                 text_column: Optional[Union[str, List]] = None):
         """Creates an RAITextInsights object.
 
         :param model: The model to compute RAI insights for.
             A model that implements sklearn.predict or sklearn.predict_proba
             or function that accepts a 2d ndarray.
         :type model: object
-        :param train: The training dataset including the label column.
-        :type train: pandas.DataFrame
         :param test: The test dataset including the label column.
         :type test: pandas.DataFrame
         :param target_column: The name of the label column or list of columns.
@@ -82,27 +141,41 @@ class RAITextInsights(RAIBaseInsights):
         :param maximum_rows_for_test: Limit on size of test data
             (for performance reasons)
         :type maximum_rows_for_test: int
+        :param feature_metadata: Feature metadata for the dataset
+            to identify different kinds of features.
+        :type feature_metadata: Optional[FeatureMetadata]
+        :param text_column: The name of the optional text column(s).
+            If not provided, and there is additional feature metadata, then
+            an exception will be raised.
+        :type text_column: str or list[str]
         """
         # drop index as this can cause issues later like when copying
         # target column below from test dataset to _ext_test_df
-        train = train.reset_index(drop=True)
         test = test.reset_index(drop=True)
+        if feature_metadata is None:
+            # initialize to avoid having to keep checking if it is None
+            feature_metadata = FeatureMetadata()
+        feature_metadata = _add_extra_metadata_features(
+            task_type, feature_metadata)
+        self._text_column = text_column
+        self._feature_metadata = feature_metadata
         self._wrapped_model = wrap_model(model, test, task_type)
         self._validate_rai_insights_input_parameters(
-            model=self._wrapped_model, train=train, test=test,
+            model=self._wrapped_model, test=test,
             target_column=target_column, task_type=task_type,
             classes=classes,
             serializer=serializer,
-            maximum_rows_for_test=maximum_rows_for_test)
+            maximum_rows_for_test=maximum_rows_for_test,
+            text_column=self._text_column)
         self._classes = RAITextInsights._get_classes(
             task_type=task_type,
-            train=train,
+            test=test,
             target_column=target_column,
             classes=classes
         )
-        ext_test, ext_features = extract_features(test,
-                                                  target_column,
-                                                  task_type)
+        ext_test, ext_features = extract_features(
+            test, target_column, task_type,
+            self._feature_metadata.dropped_features)
         self._ext_test = ext_test
         self._ext_features = ext_features
         self._ext_test_df = pd.DataFrame(ext_test, columns=ext_features)
@@ -110,7 +183,7 @@ class RAITextInsights(RAIBaseInsights):
         self.predict_output = None
 
         super(RAITextInsights, self).__init__(
-            model, train, test, target_column, task_type,
+            model, None, test, target_column, task_type,
             serializer)
         self._initialize_managers()
 
@@ -120,21 +193,22 @@ class RAITextInsights(RAIBaseInsights):
         Initializes the explainer and error analysis managers.
         """
         self._explainer_manager = ExplainerManager(
-            self.model, self.train, self.test,
+            self.model, self.test,
             self.target_column,
             self.task_type,
             self._classes)
         self._error_analysis_manager = ErrorAnalysisManager(
             self._wrapped_model, self.test, self._ext_test_df,
-            self.target_column, self.task_type, self._classes)
+            self.target_column, self._text_column, self.task_type,
+            self._classes, self._feature_metadata.categorical_features)
         self._managers = [self._explainer_manager,
                           self._error_analysis_manager]
 
     @staticmethod
-    def _get_classes(task_type, train, target_column, classes):
+    def _get_classes(task_type, test, target_column, classes):
         if task_type == ModelTask.TEXT_CLASSIFICATION:
             if classes is None:
-                classes = train[target_column].unique()
+                classes = test[target_column].unique()
                 # sort the classes after calling unique in numeric case
                 classes.sort()
                 return classes
@@ -168,14 +242,13 @@ class RAITextInsights(RAIBaseInsights):
             raise UserConfigValidationException(
                 'The serializer should be serializable via pickle')
 
-    def _validate_model(self, model: Any, train: pd.DataFrame, test: pd.DataFrame,
-                        target_column: Union[str, List], task_type: str):
+    def _validate_model(self, model: Any, test: pd.DataFrame,
+                        target_column: Union[str, List], task_type: str,
+                        text_column: Optional[Union[str, List]]):
         """Validate the model.
 
         :param model: The model to validate.
         :type model: object
-        :param train: The training dataset including the label column.
-        :type train: pandas.DataFrame
         :param test: The test dataset including the label column.
         :type test: pandas.DataFrame
         :param target_column: The name of the label column or list of columns.
@@ -184,20 +257,22 @@ class RAITextInsights(RAIBaseInsights):
         :param task_type: The task to run, can be `classification` or
             `regression`.
         :type task_type: str
+        :param text_column: The name of the optional text column(s).
+            If not provided, and there is additional feature metadata, then
+            an exception will be raised.
+        :type text_column: str or list[str]
         """
         if not isinstance(target_column, list):
             target_column = [target_column]
-        # Pick one row from train and test data
-        small_train_data = train.iloc[0:1].drop(
-            target_column, axis=1).iloc[0]
+        # Pick one row from test data
         small_test_data = test.iloc[0:1].drop(
-            target_column, axis=1).iloc[0]
+            target_column, axis=1)
+        small_test_data = get_text_columns(small_test_data, text_column)
+        small_test_data = small_test_data.iloc[0]
         if task_type != ModelTask.QUESTION_ANSWERING:
-            small_train_data = small_train_data.tolist()
             small_test_data = small_test_data.tolist()
         # Call the model
         try:
-            model.predict(small_train_data)
             model.predict(small_test_data)
         except Exception:
             raise UserConfigValidationException(
@@ -206,19 +281,18 @@ class RAITextInsights(RAIBaseInsights):
             )
 
     def _validate_rai_insights_input_parameters(
-            self, model: Any, train: pd.DataFrame, test: pd.DataFrame,
+            self, model: Any, test: pd.DataFrame,
             target_column: Union[str, List], task_type: str,
             classes: np.ndarray,
             serializer,
-            maximum_rows_for_test: int):
+            maximum_rows_for_test: int,
+            text_column: Optional[Union[str, List]]):
         """Validate the inputs for the RAITextInsights constructor.
 
         :param model: The model to compute RAI insights for.
             A model that implements sklearn.predict or sklearn.predict_proba
             or function that accepts a 2d ndarray.
         :type model: object
-        :param train: The training dataset including the label column.
-        :type train: pandas.DataFrame
         :param test: The test dataset including the label column.
         :type test: pandas.DataFrame
         :param target_column: The name of the label column or list of columns.
@@ -261,12 +335,11 @@ class RAITextInsights(RAIBaseInsights):
         if serializer is not None:
             self._validate_serializer(serializer)
 
-        train_is_pd = isinstance(train, pd.DataFrame)
         test_is_pd = isinstance(test, pd.DataFrame)
-        if not train_is_pd or not test_is_pd:
+        if not test_is_pd:
             raise UserConfigValidationException(
-                "Unsupported data type for either train or test. "
-                "Expecting pandas DataFrame for train and test."
+                "Unsupported data type for test dataset. "
+                "Expecting pandas DataFrame."
             )
 
         if test.shape[0] > maximum_rows_for_test:
@@ -279,42 +352,39 @@ class RAITextInsights(RAIBaseInsights):
                     test.shape[0], maximum_rows_for_test)
             )
 
-        if len(set(train.columns) - set(test.columns)) != 0 or \
-                len(set(test.columns) - set(train.columns)):
-            raise UserConfigValidationException(
-                'The features in train and test data do not match')
-
         if task_type == ModelTask.MULTILABEL_TEXT_CLASSIFICATION.value:
             if not isinstance(target_column, list):
                 raise UserConfigValidationException(
                     'The target_column should be a list for multilabel '
                     'classification')
-            # check all target columns are present in train and test
+            # check all target columns are present in test
             target_columns_set = set(target_column)
-            if not target_columns_set.issubset(set(train.columns)):
-                raise UserConfigValidationException(
-                    'The list of target_column(s) should be in train data')
             if not target_columns_set.issubset(set(test.columns)):
                 raise UserConfigValidationException(
                     'The list of target_column(s) should be in test data')
         else:
-            if target_column not in list(train.columns) or \
-                    target_column not in list(test.columns):
+            if target_column not in list(test.columns):
                 raise UserConfigValidationException(
-                    'Target name {0} not present in train/test data'.format(
+                    'Target name {0} not present in test data'.format(
                         target_column)
                 )
 
-        if classes is not None and task_type == \
-                ModelTask.TEXT_CLASSIFICATION:
-            if len(set(train[target_column].unique()) -
-                    set(classes)) != 0 or \
-                    len(set(classes) -
-                        set(train[target_column].unique())) != 0:
-                raise UserConfigValidationException(
-                    'The train labels and distinct values in '
-                    'target (train data) do not match')
+        if text_column:
+            if task_type == ModelTask.QUESTION_ANSWERING.value:
+                if not isinstance(text_column, list):
+                    raise UserConfigValidationException(
+                        'The text_column should be a list for question ' +
+                        'answering')
+                text_columns_set = set(text_column)
+                if not text_columns_set.issubset(set(test.columns)):
+                    raise UserConfigValidationException(
+                        'The list of text_column(s) should be in test data')
+            else:
+                if text_column not in test.columns:
+                    raise UserConfigValidationException(
+                        'The text_column should be in test data')
 
+        if classes is not None and task_type == ModelTask.TEXT_CLASSIFICATION:
             if len(set(test[target_column].unique()) -
                     set(classes)) != 0:
                 raise UserConfigValidationException(
@@ -322,8 +392,8 @@ class RAITextInsights(RAIBaseInsights):
                     'distinct values in target (test data)')
 
         if model is not None:
-            self._validate_model(model, train, test, target_column,
-                                 task_type)
+            self._validate_model(model, test, target_column,
+                                 task_type, text_column)
 
     def get_filtered_test_data(self, filters, composite_filters,
                                include_original_columns_only=False,
@@ -413,8 +483,7 @@ class RAITextInsights(RAIBaseInsights):
         self._save_ext_data(path)
 
     def _save_ext_data(self, path):
-        """Save the copy of raw data (train and test sets) and
-           their related metadata.
+        """Save the copy of raw data and their related metadata.
 
         :param path: The directory path to save the RAIBaseInsights to.
         :type path: str
@@ -429,25 +498,27 @@ class RAITextInsights(RAIBaseInsights):
         with open(data_path, 'w') as file:
             json.dump(data, file, default=serialize_json_safe)
 
-    def _get_test_without_target(self, is_classification_task):
-        """Get the test data without the target column.
+    def _get_test_text_data(self, is_classification_task):
+        """Get the test data without the target and metadata columns.
 
         :param is_classification_task: Whether the task is a
             classification task.
         :type is_classification_task: bool
-        :return: The test data without the target column.
-        :rtype: pandas.DataFrame
+        :return: The test data without the target and metadata columns.
+        :rtype: pandas.DataFrame or list[str]
         """
         if is_classification_task:
             target_column = self.target_column
             if not isinstance(target_column, list):
                 target_column = [target_column]
-            dataset = self.test.drop(
-                target_column, axis=1).iloc[:, 0].tolist()
+            dataset = self.test.drop(target_column, axis=1)
         elif self.task_type == ModelTask.QUESTION_ANSWERING:
             dataset = self.test.drop([self.target_column], axis=1)
         else:
             raise ValueError("Unknown task type: {}".format(self.task_type))
+        dataset = get_text_columns(dataset, self._text_column)
+        if is_classification_task:
+            dataset = dataset.iloc[:, 0].tolist()
         return dataset
 
     @property
@@ -464,16 +535,17 @@ class RAITextInsights(RAIBaseInsights):
     def _get_dataset(self):
         dashboard_dataset = Dataset()
         tasktype = self.task_type
-        is_classification_task = self._is_classification_task
-        if is_classification_task:
-            tasktype = RAIModelTask.CLASSIFICATION
         if isinstance(tasktype, Enum):
             tasktype = tasktype.value
         dashboard_dataset.task_type = tasktype
-        dashboard_dataset.categorical_features = []
+        categorical_features = self._feature_metadata.categorical_features
+        if categorical_features is None:
+            categorical_features = []
+        dashboard_dataset.categorical_features = categorical_features
         dashboard_dataset.class_names = convert_to_list(
             self._classes)
-        dataset = self._get_test_without_target(is_classification_task)
+        is_classification_task = self._is_classification_task
+        dataset = self._get_test_text_data(is_classification_task)
 
         predicted_y = None
         if dataset is not None and self._wrapped_model is not None:
@@ -561,7 +633,7 @@ class RAITextInsights(RAIBaseInsights):
         if self.model is None:
             return
         is_classification_task = self._is_classification_task
-        test_without_target_column = self._get_test_without_target(
+        test_without_target_column = self._get_test_text_data(
             is_classification_task)
         predict_output = self._wrapped_model.predict(
             test_without_target_column)
@@ -585,10 +657,13 @@ class RAITextInsights(RAIBaseInsights):
         """
         top_dir = Path(path)
         classes = convert_to_list(self._classes)
+        feature_metadata_dict = self._feature_metadata.to_dict()
         meta = {
             _TARGET_COLUMN: self.target_column,
             _TASK_TYPE: self.task_type,
-            _CLASSES: classes
+            _CLASSES: classes,
+            _FEATURE_METADATA: feature_metadata_dict,
+            _TEXT_COLUMN: self._text_column
         }
         with open(top_dir / _META_JSON, 'w') as file:
             json.dump(meta, file)
@@ -609,18 +684,28 @@ class RAITextInsights(RAIBaseInsights):
         inst.__dict__[_TARGET_COLUMN] = meta[_TARGET_COLUMN]
         inst.__dict__[_TASK_TYPE] = meta[_TASK_TYPE]
         inst.__dict__[_PREDICT_OUTPUT] = None
-        classes = None
-        if _TRAIN_LABELS in meta:
-            classes = meta[_TRAIN_LABELS]
-        else:
-            classes = meta[_CLASSES]
+        text_column = None
+        if _TEXT_COLUMN in meta:
+            text_column = meta[_TEXT_COLUMN]
+        inst.__dict__[_TEXT_COLUMN] = text_column
+        classes = meta[_CLASSES]
 
         inst.__dict__['_' + _CLASSES] = RAITextInsights._get_classes(
             task_type=meta[_TASK_TYPE],
-            train=inst.__dict__[_TRAIN],
+            test=inst.__dict__[_TEST],
             target_column=meta[_TARGET_COLUMN],
             classes=classes
         )
+
+        if (Metadata.FEATURE_METADATA not in meta or
+                meta[Metadata.FEATURE_METADATA] is None):
+            inst.__dict__['_' + Metadata.FEATURE_METADATA] = FeatureMetadata()
+        else:
+            feature_metadata_dict = meta[Metadata.FEATURE_METADATA]
+            feature_metadata = _feature_metadata_from_dict(
+                feature_metadata_dict)
+            inst.__dict__['_' + Metadata.FEATURE_METADATA] = feature_metadata
+
         # load the extracted features as part of metadata
         RAITextInsights._load_ext_data(inst, path)
 
@@ -659,7 +744,6 @@ class RAITextInsights(RAIBaseInsights):
         # create the RAITextInsights without any properties using the __new__
         # function, similar to pickle
         inst = RAITextInsights.__new__(RAITextInsights)
-        inst.categorical_features = []
 
         manager_map = {
             ManagerNames.EXPLAINER: ExplainerManager,
@@ -671,3 +755,101 @@ class RAITextInsights(RAIBaseInsights):
                               RAITextInsights._load_metadata)
         inst._wrapped_model = wrap_model(inst.model, inst.test, inst.task_type)
         return inst
+
+    def normalize_text(self, s):
+        """Normalize the text.
+
+        Removing articles and punctuation, and standardizing whitespace
+        are all typical text processing steps.
+
+        :param s: The text to normalize.
+        :type s: str
+        :return: The normalized text.
+        :rtype: str
+        """
+        import re
+        import string
+
+        def remove_articles(text):
+            regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
+            return re.sub(regex, " ", text)
+
+        def white_space_fix(text):
+            return " ".join(text.split())
+
+        def remove_punc(text):
+            exclude = set(string.punctuation)
+            return "".join(ch for ch in text if ch not in exclude)
+
+        def lower(text):
+            return text.lower()
+
+        return white_space_fix(remove_articles(remove_punc(lower(s))))
+
+    def compute_f1(self, prediction, truth):
+        pred_tokens = self.normalize_text(prediction).split()
+        truth_tokens = self.normalize_text(truth).split()
+
+        # if either the prediction or the truth is no-answer
+        # then f1 = 1 if they agree, 0 otherwise
+        if len(pred_tokens) == 0 or len(truth_tokens) == 0:
+            return int(pred_tokens == truth_tokens)
+
+        common_tokens = set(pred_tokens) & set(truth_tokens)
+
+        # if there are no common tokens then f1 = 0
+        if len(common_tokens) == 0:
+            return 0
+
+        prec = len(common_tokens) / len(pred_tokens)
+        rec = len(common_tokens) / len(truth_tokens)
+
+        return 2 * (prec * rec) / (prec + rec)
+
+    def compute_question_answering_metrics(
+        self,
+        selection_indexes,
+        question_answering_cache
+    ):
+        dashboard_dataset = self.get_data().dataset
+        true_y = dashboard_dataset.true_y
+        predicted_y = dashboard_dataset.predicted_y
+        all_cohort_metrics = []
+        for cohort_indices in selection_indexes:
+            key = ','.join([str(cid) for cid in cohort_indices])
+            if key in question_answering_cache:
+                all_cohort_metrics.append(question_answering_cache[key])
+                continue
+            true_y_cohort = [true_y[cohort_index] for cohort_index
+                             in cohort_indices]
+            predicted_y_cohort = [predicted_y[cohort_index] for cohort_index
+                                  in cohort_indices]
+            f1_score = []
+            for cohort_index in cohort_indices:
+                f1_score.append(self.compute_f1(predicted_y[cohort_index],
+                                                true_y[cohort_index]))
+            try:
+                exact_match = evaluate.load('exact_match')
+                exact_match_results = exact_match.compute(
+                    predictions=predicted_y_cohort, references=true_y_cohort)
+                rouge = evaluate.load('rouge')
+                rouge_results = rouge.compute(
+                    predictions=predicted_y_cohort, references=true_y_cohort)
+                bleu = evaluate.load('bleu')
+                bleu_results = bleu.compute(
+                    predictions=predicted_y_cohort, references=true_y_cohort)
+                meteor = evaluate.load('meteor')
+                meteor_results = meteor.compute(
+                    predictions=predicted_y_cohort, references=true_y_cohort)
+                bert_score = evaluate.load('bertscore')
+                bert_score_results = bert_score.compute(
+                    predictions=predicted_y_cohort, references=true_y_cohort,
+                    model_type="distilbert-base-uncased")
+                bert_f1_score = np.mean(bert_score_results['f1'])
+                all_cohort_metrics.append(
+                    [exact_match_results['exact_match'], np.mean(f1_score),
+                     meteor_results['meteor'], bleu_results['bleu'],
+                     bert_f1_score, rouge_results['rougeL']])
+            except ValueError:
+                all_cohort_metrics.append([0, 0, 0, 0, 0, 0])
+        return all_cohort_metrics
